/* $Id: dmcalibinit.cpp,v 1.15 2004/07/13 16:26:21 dmoroz Exp $*/

#include <cv.h>
#include <_cv.h>
#include "dmcalibinit.h"
#include "dmcalibinternal.h"

//=====================================================================================
// Implementation for the enhanced calibration object detection
//=====================================================================================

#define PROCESSED_CORNER    1
#define MAX_CONTOUR_APPROX  7


typedef struct CvContourEx
{
    CV_CONTOUR_FIELDS()
    int counter;
}
CvContourEx;

IplImage *g_ppDebugImages[DEBUG_IMG_NUM] = {NULL, NULL, NULL, NULL, NULL, NULL};


//=====================================================================================

double DQuadInfo::GetMinDim()
{
  double dMin = CV_SQR(m_pCorners[0]->m_vPoint.x - m_pCorners[3]->m_vPoint.x) + 
                CV_SQR(m_pCorners[0]->m_vPoint.y - m_pCorners[3]->m_vPoint.y),
         dTemp;

  for (int i = 0; i < 3; i++)
  {
    dTemp = CV_SQR(m_pCorners[i]->m_vPoint.x - m_pCorners[i + 1]->m_vPoint.x) + 
            CV_SQR(m_pCorners[i]->m_vPoint.y - m_pCorners[i + 1]->m_vPoint.y);

    dMin = CV_MIN(dMin, dTemp);
  }  

  return dMin;
}

//=====================================================================================

int compare( const void *arg1, const void *arg2 )
{
   /* Compare all of both strings: */
   return ( * ( double* ) arg1 > * ( double* ) arg2 ) - ( * ( double* ) arg1 < * ( double* ) arg2 );
}

//=====================================================================================

CV_IMPL
int dmFindChessBoardCornerGuesses( const void* arr, void* thresharr,
                                   CvMemStorage * storage,
                                   CvSize etalon_size, CvPoint2D32f * corners,
                                   int *corner_count, int in_bAdaptiveThresh)
{
    CvStatus result = CV_NO_ERR;
    int iQuadCount,
        iCornerCount,
        iGroups,
        iCount,
        i;

    DQuadInfo   *pQuads = NULL,
               **ppQuadGroup = NULL;
    // Corners of internal part of object
    DCornerInfo *pCorners = NULL,
               **ppCornerGroup = NULL;


    CV_FUNCNAME( "dmFindChessBoardCornerGuesses" );

    __BEGIN__;

    result = CV_NOTDEFINED_ERR;

    CvMat  stub, *img = (CvMat*)arr;
    CvMat  thstub, *thresh = (CvMat*)thresharr;
    CvSize size;

    CV_CALL( img    = cvGetMat( img, &stub ));
    CV_CALL( thresh = cvGetMat( thresh, &thstub ));

    if( CV_MAT_TYPE( img->type ) != CV_8UC1 ||
        CV_MAT_TYPE( thresh->type ) != CV_8UC1 )
        CV_ERROR( CV_BadDepth, icvUnsupportedFormat );

    if( !CV_ARE_SIZES_EQ( img, thresh ))
        CV_ERROR( CV_StsUnmatchedSizes, "" );
 
    size = icvGetMatSize( img );

    #ifdef DEBUG_DRAW
    ReleaseTmpImages();
    for (i = 0; i < DEBUG_IMG_NUM; i++)
    {    
      g_ppDebugImages[i] = cvCreateImage( size, IPL_DEPTH_8U, 1 );
      cvCopy(img, g_ppDebugImages[i]);
    }
    
    #endif

    ThresholdImage(img, thresh, in_bAdaptiveThresh,
                   (CvMat *)g_ppDebugImages[5]);

    iQuadCount = GenerateQuads(&pQuads, &pCorners, storage,
                               thresh, (CvMat *)g_ppDebugImages[0]);
  
    // Rem - corners are quads * 4
    iCornerCount = iQuadCount * 4;

    if (iQuadCount < 1)
    {
      result = CV_NOTDEFINED_ERR;
      EXIT;
    }

    //  Find quad's neibors
    FindQuadNeibors(pQuads, iQuadCount, pCorners, iCornerCount, (CvMat *)g_ppDebugImages[1]);  

    iGroups = 0;

    ppQuadGroup = (DQuadInfo **)cvAlloc(sizeof(DQuadInfo *) * iQuadCount);
    ppCornerGroup = (DCornerInfo **)cvAlloc(sizeof(DCornerInfo *) * iCornerCount);

    do
    {

      //    
      iCount = FindConnectedQuads(pQuads, iQuadCount,
                                  ppQuadGroup, iGroups,
                                  (CvMat *)g_ppDebugImages[2]);

      // "" 
      if (CheckQuadGroup(ppQuadGroup, iCount, ppCornerGroup, etalon_size, 
                         (CvMat *)g_ppDebugImages[3]))
      {
        // Copy corners to output array
        *corner_count = (etalon_size.width) * (etalon_size.height);

        for (i = 0; i < *corner_count; i++)
          corners[i] = ppCornerGroup[i]->m_vPoint;

        result = CV_OK;
        EXIT;
      }

      iGroups++;
    }
    while(iCount)

    __CLEANUP__;
    __END__;

    cvFree( (void **)&pQuads );
    cvFree( (void **)&pCorners );
    cvFree( (void **)&ppQuadGroup );
    cvFree( (void **)&ppCornerGroup );
    
    return result == CV_OK;
}


//=====================================================================================

void ThresholdImage(CvMat *in_pImage, CvMat *out_pThreshImage,
                    int in_bLocalAdaptive, CvMat *out_pDebugImage)
{
  CvMat *pTempImage = (CvMat *)cvClone(in_pImage);

  if (in_bLocalAdaptive)
  {
    int iRadius = in_pImage->width * 50 / 320; 

    cvSmooth( in_pImage, pTempImage, CV_BLUR, 3, 3); 

    cvDilate( pTempImage, out_pThreshImage, 0, 1 );

    iRadius = iRadius + (1 - iRadius & 1);

    // convert to binary 
    cvAdaptiveThreshold ( out_pThreshImage, out_pThreshImage, 255,  
                          CV_ADAPTIVE_THRESH_GAUSSIAN_C, CV_THRESH_BINARY, 
                          iRadius, 10);

    //cvDilate( pTempImage, out_pThreshImage, 0, 1 );
  }
  else 
  {  
    // Make dilation before the thresholding. 
    // It splits chessboard corners 
    cvDilate( in_pImage, out_pThreshImage, 0, 1 );

    double mean;
    int thresh_level; 

    // empiric threshold level 
    mean = cvMean( in_pImage );
    thresh_level = cvRound( mean - 10 );
    thresh_level = MAX( thresh_level, 10 );
 
    // convert to binary 
    cvThreshold( out_pThreshImage, out_pThreshImage, thresh_level, 255, CV_THRESH_BINARY );
  }

  //cvErode( in_pImage, out_pThreshImage, 0, 1 );

  cvRelease((void **)&pTempImage);
#ifdef DEBUG_DRAW
  if (out_pDebugImage)
    cvCopy(out_pThreshImage, out_pDebugImage);
#endif //DEBUG_DRAW
}


//=====================================================================================

int FindConnectedQuads(DQuadInfo *in_pQuadInfo, int in_iQuadNum, 
                       DQuadInfo **out_ppQuadGroup, int in_iGroup,
                       CvMat *out_pDebugImage)
{
  int i, iCount;

  // Scan the array for a first unlabeled quad
  for (i = 0; i < in_iQuadNum; i++)
  {
    if (in_pQuadInfo[i].m_iCount > 0 && in_pQuadInfo[i].m_iGroup < 0)
      break;
  }

  if (i >= in_iQuadNum)
    return 0;

  // Start very simple recursive fill algorithm
  RecursiveFill(&in_pQuadInfo[i], in_iGroup);

  // Make array of pointers to current group
  iCount = 0;

  for (i = 0; i < in_iQuadNum; i++)
  {
    if (in_pQuadInfo[i].m_iGroup == in_iGroup)
    {
      out_ppQuadGroup[iCount] = &in_pQuadInfo[i];

#ifdef DEBUG_DRAW
      if (out_pDebugImage)
        DrawQuad(out_pDebugImage, out_ppQuadGroup[iCount], 255 - (in_iGroup + 1) * 50);
#endif //DEBUG_DRAW

      iCount++;
   }
  }

  return iCount;
}

//=====================================================================================

void RecursiveFill(DQuadInfo *in_pSeed, int in_iGroup)
{
  int        i;
  DQuadInfo *pNeibor;

  // Mark current quad
  in_pSeed->m_iGroup = in_iGroup;

  // Fill all the neighbors
  for (i = 0; i < 4; i++)
  {
    if (!in_pSeed->m_pNeibors[i])
      continue;
    pNeibor = in_pSeed->m_pNeibors[i];

    if (pNeibor->m_iCount > 0 && pNeibor->m_iGroup == -1)
      RecursiveFill(pNeibor, in_iGroup);                  
  }
  return;
}

//=====================================================================================

bool CheckQuadGroup(DQuadInfo   **in_ppQuadGroup, int in_iCount,
                    DCornerInfo **out_ppCorners, CvSize in_ObjDim, 
                    CvMat *out_pDebugImage)
{
  DQuadInfo *pLeftCurQuad = NULL,
            *pRightCurQuad = NULL,
            *pTempQuad;

  DCornerInfo **ppCorners = NULL;

  int        iTemp, iLeftPoint = -1, iRightPoint= -1, i, j, k, iTemp2,
             iCornerCount = 0, iFoundCorners;
  double     dTestDist = 1e20, dTemp, dMinDist;
  bool       bTranspose = false;

  if (in_ObjDim.width % 2 + in_ObjDim.height % 2 != 1)
    return false;

  if (in_ObjDim.height % 2)
  {
    iTemp = in_ObjDim.height;
    in_ObjDim.height = in_ObjDim.width;
    in_ObjDim.width = iTemp;

    bTranspose = true;
  }

  //-------------------------------------------
  // Check the number of quads
  if (in_iCount != ((in_ObjDim.width + 1) >> 1) * (in_ObjDim.height + 1))
    return false;

  //-------------------------------------------
  // Find minimal distance between corners

  for (i = 0; i < in_iCount; i++)
  {
    double dDist;
    int    li, lj;

    if (in_ppQuadGroup[i]->m_iCount >= 2)
      continue;

    for (j = 0; j < 4; j++)
    {
      if (!in_ppQuadGroup[i]->m_pNeibors[j])
        continue;

      for (li = i + 1; li < in_iCount; li++)
      {
        for (lj = 0; lj < 4; lj++)
        {
          if (!in_ppQuadGroup[li]->m_pNeibors[lj])
            continue;
          if (in_ppQuadGroup[i]->m_pCorners[j] == in_ppQuadGroup[li]->m_pCorners[lj]) 
            continue;

          dDist = (in_ppQuadGroup[i]->m_pCorners[j]->m_vPoint.x - in_ppQuadGroup[li]->m_pCorners[lj]->m_vPoint.x) * 
                  (in_ppQuadGroup[i]->m_pCorners[j]->m_vPoint.x - in_ppQuadGroup[li]->m_pCorners[lj]->m_vPoint.x) +
                  (in_ppQuadGroup[i]->m_pCorners[j]->m_vPoint.y - in_ppQuadGroup[li]->m_pCorners[lj]->m_vPoint.y) * 
                  (in_ppQuadGroup[i]->m_pCorners[j]->m_vPoint.y - in_ppQuadGroup[li]->m_pCorners[lj]->m_vPoint.y);

          if (dTestDist > dDist)
            dTestDist = dDist;
        }
      }
    }
  }

  dTestDist = ceil(sqrt(dTestDist * 0.5));

  //-------------------------------------------
  // Find one corner quad
  // (it has 4 neibors, only one of which has 4 neigbors too)

  iTemp = 0;

  for (i = 0; i < in_iCount; i++)
  {
    iTemp2 = -1;
    if (in_ppQuadGroup[i]->m_iCount == 4)
    {
      int iLocalCount, j;

      iLocalCount = 0;
      for (j = 0; j < 4; j++)
      {
        if (in_ppQuadGroup[i]->m_pNeibors[j]->m_iCount == 4)
          iLocalCount++;
        if (in_ppQuadGroup[i]->m_pNeibors[j]->m_iCount == 1)
          iTemp2 = j;
      }

      if (iLocalCount <= 1)
      {
        iLeftPoint = iTemp2;
        pLeftCurQuad = in_ppQuadGroup[i];
        break;
      }
    }
  }

  if (!pLeftCurQuad || iLeftPoint < 0)
    return false;

#ifdef DEBUG_DRAW
  if (out_pDebugImage)
  {
    DrawQuad(out_pDebugImage, pLeftCurQuad, 0, 2);
    cvCircle(out_pDebugImage, cvPointFrom32f(pLeftCurQuad->m_pCorners[iLeftPoint]->m_vPoint), 3, 255);
  }
#endif //DEBUG_DRAW

  //-------------------------------------------
  // Find second corner quad
  // (it has 2 neibors, only one of which has 4 neibors)

  iTemp    = 0;
  dMinDist = 1e20;
  
  for (i = 0; i < in_iCount; i++)
  {
    if (in_ppQuadGroup[i]->m_iCount == 2)
    {
      int iLocalCount, j;

      iLocalCount = 0;
      for (j = 0; j < 4; j++)
      {
        if (!in_ppQuadGroup[i]->m_pNeibors[j])
          continue;

        if (in_ppQuadGroup[i]->m_pNeibors[j]->m_iCount == 4)
          iLocalCount++;
      }

      if (iLocalCount == 1)
      {
        for (j = 0; j < 4; j++)
        {
          if (in_ppQuadGroup[i]->m_pNeibors[j] &&
              in_ppQuadGroup[i]->m_pNeibors[j]->m_iCount == 2)
          {
            dTemp = CheckCornersCloseToLine(in_ppQuadGroup, in_iCount, 
                                            pLeftCurQuad->m_pCorners[iLeftPoint]->m_vPoint,
                                            in_ppQuadGroup[i]->m_pCorners[j]->m_vPoint,
                                            (int)dTestDist, in_ObjDim.width);
            if (dTemp < dMinDist)
            {
              dMinDist = dTemp;
              pRightCurQuad = in_ppQuadGroup[i];
              iRightPoint = j;
            }
            break;
          }
        }
      }
    }
  }

  if (!pRightCurQuad || iRightPoint < 0)
    return false; 

  //-------------------------------------------
  // Pick up two corner points
  
#ifdef DEBUG_DRAW
  if (out_pDebugImage)
  {
    DrawQuad(out_pDebugImage, pRightCurQuad, 255, 2);
    cvCircle(out_pDebugImage, cvPointFrom32f(pRightCurQuad->m_pCorners[iRightPoint]->m_vPoint), 3, 255);
  }
#endif //DEBUG_DRAW

  ppCorners = (DCornerInfo **) cvAlloc(sizeof(DCornerInfo *) * in_ObjDim.height * in_ObjDim.width);

  //-------------------------------------------
  //   
  do
  {
    //        
    iFoundCorners = FindCornersCloseToLine(in_ppQuadGroup, in_iCount, 
                                           pLeftCurQuad->m_pCorners[iLeftPoint]->m_vPoint,
                                           pRightCurQuad->m_pCorners[iRightPoint]->m_vPoint,
                                           &ppCorners[iCornerCount], (int)dTestDist);


#ifdef DEBUG_DRAW
    for (int lk = 0; (lk < iFoundCorners) && (out_pDebugImage); lk++)
      cvCircle(out_pDebugImage, cvPointFrom32f(ppCorners[iCornerCount + lk]->m_vPoint), 5, 0);
#endif //DEBUG_DRAW

    if (iFoundCorners != in_ObjDim.width)
      break;  

    iCornerCount += iFoundCorners;

    //   
    //   -      ,      
    //          

    iTemp = 1;
    do
    {
      for (j = 0; j < 4; j++)
      {
        if (pLeftCurQuad->m_pNeibors[j]->m_iCount == iTemp &&
            !pLeftCurQuad->m_pCorners[j]->m_iDoneFlag)
        {
          iLeftPoint = j;
          break;
        }
      }
      iTemp++;
    }
    while (j >= 4 && iTemp < 3);

    if (j >= 4)
      break;

    //   -      ,     
    //            
    // 

    do
    {
      for (j = 0; j < 4; j++)
      {
        if (pRightCurQuad->m_pNeibors[j] &&
            pRightCurQuad->m_pNeibors[j]->m_iCount > 0)
        {
          if (!pRightCurQuad->m_pCorners[j]->m_iDoneFlag)
          {
            iRightPoint = j;
            break;
          }
        }
      }

      // We failed to find next square with 4 neihbors - this can happen only at start
      if (j >= 4)
        //     - , restart cycle
        pRightCurQuad = pRightCurQuad->m_pNeibors[iRightPoint];
    }
    while (j >= 4);


#ifdef DEBUG_DRAW
    if (out_pDebugImage)
    {
      cvCircle(out_pDebugImage, cvPointFrom32f(pLeftCurQuad->m_pCorners[iLeftPoint]->m_vPoint), 3, 255);
      cvCircle(out_pDebugImage, cvPointFrom32f(pRightCurQuad->m_pCorners[iRightPoint]->m_vPoint), 3, 255);
    }
#endif //DEBUG_DRAW

    //        
    iFoundCorners = FindCornersCloseToLine(in_ppQuadGroup, in_iCount, 
                                           pLeftCurQuad->m_pCorners[iLeftPoint]->m_vPoint,
                                           pRightCurQuad->m_pCorners[iRightPoint]->m_vPoint,
                                           &ppCorners[iCornerCount], (int)dTestDist);


#ifdef DEBUG_DRAW
    for (lk = 0; (lk < iFoundCorners) && (out_pDebugImage); lk++)
      cvCircle(out_pDebugImage, cvPointFrom32f(ppCorners[iCornerCount + lk]->m_vPoint), 5, 0);
#endif //DEBUG_DRAW

    if (iFoundCorners != in_ObjDim.width)
      break;

    iCornerCount += iFoundCorners;

    //   -     iLeftPoint    
    //      
    pTempQuad = pLeftCurQuad->m_pNeibors[iLeftPoint];
    for (j = 0; j < 4; j++)
    {
      if (pTempQuad->m_pNeibors[j] &&
          pTempQuad->m_pNeibors[j]->m_iCount == 4 &&
          pTempQuad->m_pNeibors[j] != pLeftCurQuad)
      {
        pLeftCurQuad = pTempQuad->m_pNeibors[j];

        for (k = 0; k < 4; k++)
        {
          if (pLeftCurQuad->m_pNeibors[k] == pTempQuad)
          {
            iLeftPoint = k;
            break;
          }
        }
        break;
      }
    }

    if (j >= 4 || k >= 4)
      // Failed to find next left quad and/or left point
      break;

    //   -  ,      
    pTempQuad = pRightCurQuad->m_pNeibors[iRightPoint];
    for (j = 0; j < 4; j++)
    {
      if (pTempQuad->m_pNeibors[j] &&
          pTempQuad->m_pNeibors[j]->m_iCount == 2 &&
          pTempQuad->m_pNeibors[j] != pRightCurQuad)
      {
        pRightCurQuad = pTempQuad->m_pNeibors[j];

        for (k = 0; k < 4; k++)
        {
          if (pRightCurQuad->m_pNeibors[k] == pTempQuad)
          {
            iRightPoint = k;
            break;
          }
        }
        break;
      }
    }

    if (j >= 4 || k >= 4)
      // Failed to find next right quad and/or left point
      break;


#ifdef DEBUG_DRAW
    if (out_pDebugImage)
    {
      DrawQuad(out_pDebugImage, pLeftCurQuad, 0, 1);
      DrawQuad(out_pDebugImage, pRightCurQuad, 255, 1);

      cvCircle(out_pDebugImage, cvPointFrom32f(pLeftCurQuad->m_pCorners[iLeftPoint]->m_vPoint), 3, 255);
      cvCircle(out_pDebugImage, cvPointFrom32f(pRightCurQuad->m_pCorners[iRightPoint]->m_vPoint), 3, 255);
    }
#endif //DEBUG_DRAW

  }
  while (1);

  if (iCornerCount == (in_ObjDim.width) * (in_ObjDim.height))
  {

    // calculate vector product, if it's positive,
    // reverse all the rows 
    if( (ppCorners[in_ObjDim.width - 1]->m_vPoint.x - ppCorners[0]->m_vPoint.x) *
        (ppCorners[in_ObjDim.width]->m_vPoint.y - ppCorners[in_ObjDim.width - 1]->m_vPoint.y) -
        (ppCorners[in_ObjDim.width - 1]->m_vPoint.y - ppCorners[0]->m_vPoint.y) *
        (ppCorners[in_ObjDim.width]->m_vPoint.x - ppCorners[in_ObjDim.width - 1]->m_vPoint.x) > 0 )
    {
        for( i = 0; i < in_ObjDim.width; i++ )
            for( j = 0; j < in_ObjDim.height / 2; j++ )
            {
                DCornerInfo *pTemp = ppCorners[j * in_ObjDim.width + i];

                ppCorners[j * in_ObjDim.width + i] =
                    ppCorners[(in_ObjDim.height - 1 - j) * in_ObjDim.width + i];
                ppCorners[(in_ObjDim.height - 1 - j) * in_ObjDim.width + i] = pTemp;
            }
    }

    // Transpose the corners
    if (bTranspose)
    {
      for (i = 0; i < in_ObjDim.height; i++)
        for (j = 0; j < in_ObjDim.width; j++)
          out_ppCorners[i + j * in_ObjDim.height] = ppCorners[j + i * in_ObjDim.width];
    }
    else
    {
      for (i = 0; i < in_ObjDim.height; i++)
        for (j = 0; j < in_ObjDim.width; j++)
          out_ppCorners[j + i * in_ObjDim.width] = ppCorners[j + i * in_ObjDim.width];
    }

    cvFree((void **)&ppCorners);
    return true;
  }
  else
  {
    cvFree((void **)&ppCorners);
    return false;
  }
}

//=====================================================================================

int FindCornersCloseToLine(DQuadInfo **in_ppQuadGroup, int in_iCount,
                           CvPoint2D32f in_vStart, CvPoint2D32f in_vFinish, 
                           DCornerInfo **io_ppCorners, int in_iTestDist)
{
  int          i, j, k, iCount = 0, iDenom, iX, iY;
  DCornerInfo *pTempCorner,
              *pFoundCorners[50];
  CvPoint      vStartPt, vCurPt, vTempPt;

  iX = (int)(in_vStart.x - in_vFinish.x);
  iY = (int)(in_vStart.y - in_vFinish.y);
  iDenom = iX * iX + iY * iY;

  vStartPt = cvPointFrom32f(in_vStart);

  // Iterate through all quads in group
  for( i = 0; i < in_iCount; i++ )
  {
    // Through each point
    for (j = 0; j < 4; j++)
    {
      int num;
      
      pTempCorner = in_ppQuadGroup[i]->m_pCorners[j];

      if (pTempCorner->m_iDoneFlag || !in_ppQuadGroup[i]->m_pNeibors[j])
        continue;

      num = (int)((pTempCorner->m_vPoint.x - in_vStart.x) * iY -
                  (pTempCorner->m_vPoint.y - in_vStart.y) * iX);

      if( ((int64) num) * num < ((int64) iDenom) * in_iTestDist )
      {
        int iCurDist;

        vCurPt = cvPointFrom32f(pTempCorner->m_vPoint);
        iCurDist = (vStartPt.x - vCurPt.x) * (vStartPt.x - vCurPt.x) +
                   (vStartPt.y - vCurPt.y) * (vStartPt.y - vCurPt.y);

        // Add it sorted by the distance from start point
        for (k = 0; k < iCount; k++)
        {
          int iDist;

          vTempPt = cvPointFrom32f(pFoundCorners[k]->m_vPoint);
          iDist = (vStartPt.x - vTempPt.x) * (vStartPt.x - vTempPt.x) +
                  (vStartPt.y - vTempPt.y) * (vStartPt.y - vTempPt.y);

          if (iDist > iCurDist)
            break;
        }

        if (k < iCount)        
          memmove(&pFoundCorners[k + 1], &pFoundCorners[k], (iCount - k) * sizeof(DCornerInfo *));        

        pFoundCorners[k] = pTempCorner;
        pFoundCorners[k]->m_iDoneFlag = PROCESSED_CORNER;
        iCount++;
      }
    }
  }

  for (k = 0; k < iCount; k++)
    io_ppCorners[k] = pFoundCorners[k];

  return iCount;
}


//=====================================================================================

double CheckCornersCloseToLine(DQuadInfo **in_ppQuadGroup, int in_iCount,
                               CvPoint2D32f in_vStart, CvPoint2D32f in_vFinish, 
                               int in_iTestDist, int in_iExpected)
{
  int          i, j, iCount = 0, iX, iY;
  int64        iMaxNum = 0, iDenom, iNum = 0;

  DCornerInfo *pTempCorner,
              *pFoundCorners[50];
  CvPoint      vStartPt;

  assert(in_iExpected < 50);

  iX = (int)(in_vStart.x - in_vFinish.x);
  iY = (int)(in_vStart.y - in_vFinish.y);
  iDenom = (int64)(iX * iX + iY * iY);

  vStartPt = cvPointFrom32f(in_vStart);

  // Iterate through all quads in group
  for( i = 0; i < in_iCount; i++ )
  {
    // Through each point
    for (j = 0; j < 4; j++)
    {      
      pTempCorner = in_ppQuadGroup[i]->m_pCorners[j];

      if (pTempCorner->m_iDoneFlag || !in_ppQuadGroup[i]->m_pNeibors[j])
        continue;

      iNum = (int64)((pTempCorner->m_vPoint.x - in_vStart.x) * iY -
                     (pTempCorner->m_vPoint.y - in_vStart.y) * iX);

      if( iNum * iNum < iDenom * in_iTestDist )
      {
        if (iNum * iNum > iMaxNum)
          iMaxNum = iNum * iNum;

        pTempCorner->m_iDoneFlag = PROCESSED_CORNER;
        pFoundCorners[iCount] = pTempCorner;
        iCount++;
      }
    }
  }

  for (i = 0; i < iCount; i++)
    pFoundCorners[i]->m_iDoneFlag = 0;

  if (iCount != in_iExpected)
    return 1e100;
  else
    return ((double)iMaxNum) / iDenom;
}

//=====================================================================================

bool IsLinesCross(int64 x1, int64 y1, int64 x2, int64 y2, int64 x3, int64 y3, int64 x4, int64 y4)
{
  int64 maxx1 = CV_MAX(x1, x2), maxy1 = CV_MAX(y1, y2);
  int64 minx1 = CV_MIN(x1, x2), miny1 = CV_MIN(y1, y2);
  int64 maxx2 = CV_MAX(x3, x4), maxy2 = CV_MAX(y3, y4);
  int64 minx2 = CV_MIN(x3, x4), miny2 = CV_MIN(y3, y4);

  if (minx1 > maxx2 || maxx1 < minx2 || miny1 > maxy2 || maxy1 < miny2)
    return false;  // ,      ...
 

  int64 dx21 = x2 - x1, dy21 = y2 - y1; //       x  y
  int64 dx43 = x4 - x3, dy43 = y4 - y3; //       x  y
  int64 dx13 = x1 - x3, dy13 = y1 - y3;
  int64 div, mul;


  if ((div = (int64)((double)dy43 * dx21 - (double)dx43 * dy21)) == 0) 
    return false; //  ...

  if (div > 0) 
  {
    if ((mul = (int64)((double)dx43 * dy13 - (double)dy43 * dx13)) < 0 || mul > div)
      return false; //      ...
    if ((mul = (int64)((double)dx21 * dy13 - (double)dy21 * dx13)) < 0 || mul > div)
      return false; //      ...
  }
  else
  {
    if ((mul = -(int64)((double)dx43 * dy13-(double)dy43 * dx13)) < 0 || mul > -div)
      return false; //      ...
    if ((mul = -(int64)((double)dx21 * dy13-(double)dy21 * dx13)) < 0 || mul > -div)
      return false; //      ...
  }

  return true;
}


//=====================================================================================

int  FindQuadNeibors(DQuadInfo *in_pQuads, int in_iQuadsNum,
                     DCornerInfo *in_pCorners, int in_iCornersNum,
                     CvMat *out_pDebugImage)
{
  int idx, i, k, j;
  DQuadInfo   *pMinQuad;
  DCornerInfo *pMinCorner;
  double dGlobalMinDist = 1e236,
         dMinQuadDim = 1e236,
         dTemp;

  double *pMinQuadDim = (double *)cvAlloc(sizeof(double) * in_iQuadsNum);

  // Find minimum quad dimension
  for( idx = 0; idx < in_iQuadsNum; idx++ )
    pMinQuadDim[idx] = in_pQuads[idx].GetMinDim();

  qsort(pMinQuadDim, in_iQuadsNum, sizeof(double), compare);

  dMinQuadDim = pMinQuadDim[in_iQuadsNum / 2] / 4;

  // Find quad neighbors
  for( idx = 0; idx < in_iQuadsNum; idx++ )
  {

#ifdef DEBUG_DRAW
    if (out_pDebugImage)
      DrawQuad(out_pDebugImage, &in_pQuads[idx], 255, 2);
#endif //DEBUG_DRAW
    //   choose the points of the current quadrangle that are close to
    //   some points of the other quadrangles
    //   (it can happen for splitted corners (due to dilation) of the
    //   checker board). Search only in other quadrangles! 

    for( i = 0; i < 4; i++ )
    {
      CvPoint2D32f pt;

      if (in_pQuads[idx].m_pNeibors[i])
        continue;

      pt = in_pQuads[idx].m_pCorners[i]->m_vPoint;

      // Find the CLOSEST point
      double dMinDist = 1e236, dDist;
      int    iMinPoint = -1;
      bool bNotFound = true;

      dMinDist = 1e236;
      pMinQuad   = NULL;
      pMinCorner = NULL; 

      for( k = 0; k < in_iQuadsNum; k++)
      {
        if (k == idx)
          continue;

        for( j = 0; j < 4; j++ )
        {
          if (in_pQuads[k].m_pNeibors[j])
            continue;

          int dx = (int)(pt.x - in_pQuads[k].m_pCorners[j]->m_vPoint.x);
          int dy = (int)(pt.y - in_pQuads[k].m_pCorners[j]->m_vPoint.y);

          dDist = dx * dx + dy * dy;

          if (dDist < dMinDist)
          {
            iMinPoint = j; 
            pMinQuad = &in_pQuads[k];
            dMinDist = dDist;
          }
        }
      }

      if (iMinPoint >= 0 && dMinDist < dMinQuadDim)
      {
        if (!bNotFound)
          goto NextPoint;
        bNotFound = false;
      }
      else
      {
        bNotFound = true;
      }

      if (!bNotFound)
      {
        int ik;

        pMinCorner = pMinQuad->m_pCorners[iMinPoint];
        pMinCorner->m_vPoint.x = (pt.x + pMinCorner->m_vPoint.x) * 0.5f;
        pMinCorner->m_vPoint.y = (pt.y + pMinCorner->m_vPoint.y) * 0.5f;

        // Check that each corner is neighbour of different counters 
        for (ik = 0; ik < in_pQuads[idx].m_iCount; ik++)
        {
          if (in_pQuads[idx].m_pNeibors[ik] == pMinQuad)
            break;
        }
        if (ik < in_pQuads[idx].m_iCount)
          break;

        for (ik = 0; ik < pMinQuad->m_iCount; ik++)
        {
          if (pMinQuad->m_pNeibors[ik] == &in_pQuads[idx])
            goto NextPoint;
        }
        if (ik < pMinQuad->m_iCount)
          break;

        // We've found one more corner - remember it
        in_pQuads[idx].m_iCount++;
        assert(in_pQuads[idx].m_iCount <= 4);
        in_pQuads[idx].m_pNeibors[i] = pMinQuad;
        in_pQuads[idx].m_pCorners[i] = pMinCorner;

        pMinQuad->m_iCount++;
        assert(pMinQuad->m_iCount <= 4);
        pMinQuad->m_pNeibors[iMinPoint] = &in_pQuads[idx];
        assert(pMinQuad->m_pCorners[iMinPoint] == pMinCorner);


#ifdef DEBUG_DRAW
        if (out_pDebugImage)
        {
          cvCircle(out_pDebugImage, 
                   cvPointFrom32f(pMinCorner->m_vPoint), 
                   2, CV_RGB(255, 255, 255));

          cvCircle(out_pDebugImage, 
                   cvPointFrom32f(pMinCorner->m_vPoint), 
                   4, CV_RGB(0, 0, 0));
        }
#endif //DEBUG_DRAW

        //  
        dGlobalMinDist = CV_MIN(dGlobalMinDist, dMinDist);
      }
NextPoint:;
    }
  }

  return (int)dGlobalMinDist;
}

//=====================================================================================

int GenerateQuads(DQuadInfo **out_ppQuads, DCornerInfo **out_ppCorners, 
                  CvMemStorage *in_pStorage, CvMat *in_pThresh,
                  CvMat *out_pDebugImage)
{
  int iQuadCount = 0;
  CvSeq *src_contour = 0;
  CvMemStorage *storage1 = 0;
  CvSeq *root;
  CvContourEx* board = 0;
  CvContourScanner scanner;
  int min_size;
  CvSize size;
  const int min_approx_level = 2;
  const int max_approx_level = MAX_CONTOUR_APPROX;
  int quandrangles = 0,
      idx;
  CvSeqReader reader;

  if (!out_ppCorners || !out_ppQuads)
    return 0;

  size = icvGetMatSize( in_pThresh );


  // empiric bound for minimal allowed perimeter for squares 
  min_size = cvRound( size.width * size.height * .03 * 0.01 );

  // 
  //   Create temporary storages.
  //   First one will store retrived contours and the
  //   second one - approximated contours.
  storage1 = in_pStorage ? cvCreateChildMemStorage( in_pStorage ) : cvCreateMemStorage(0);
  root = cvCreateSeq( 0, sizeof(CvSeq), sizeof(CvSeq*), storage1 );

  // initialize contour retrieving routine 
  scanner = cvStartFindContours( in_pThresh, storage1, sizeof( CvContourEx ),
                                 CV_RETR_CCOMP, CV_CHAIN_APPROX_SIMPLE );

  // Debug output
  //cvSetZero(img);

  // get all the contours one by one 
  while( (src_contour = cvFindNextContour( scanner )) != 0 )
  {
      CvSeq *dst_contour = 0;
      CvRect rect = ((CvContour*)src_contour)->rect;

      // reject contours with too small perimeter 
      if( CV_IS_SEQ_HOLE( src_contour ) &&
          rect.width *rect.height >= min_size )
      {
          int approx_level;

          for( approx_level = min_approx_level;
               approx_level <= max_approx_level;
               approx_level++ )
          {
              dst_contour = cvApproxPoly( src_contour, sizeof( CvContour ),
                                          storage1, CV_POLY_APPROX_DP,
                                          (float)approx_level );

              if( dst_contour->total == 4 )
                  break;
          }

          quandrangles += dst_contour->total == 4;

          // reject non-quadrangles 
          if( dst_contour->total == 4 && cvCheckContourConvexity( dst_contour ))
          {
              CvPoint pt[4];
              int i;
              double d1, d2, p = cvContourPerimeter(dst_contour);
              double dx, dy;

              for( i = 0; i < 4; i++ )
                  pt[i] = *(CvPoint*)cvGetSeqElem( dst_contour, i );

              dx = pt[0].x - pt[2].x;
              dy = pt[0].y - pt[2].y;
              d1 = sqrt( dx*dx + dy*dy );

              dx = pt[1].x - pt[3].x;
              dy = pt[1].y - pt[3].y;
              d2 = sqrt( dx*dx + dy*dy );

#ifdef DEBUG_DRAW
              if (out_pDebugImage)
                cvDrawContours(out_pDebugImage, dst_contour,
                               0, 10, 0, 1);
#endif //DEBUG_DRAW

              if( d1 >= 0.15 * p && d2 >= 0.15 * p )
              {
                  CvContourEx* parent = (CvContourEx*)(src_contour->v_prev);
                  parent->counter++;
                  if( !board || board->counter < parent->counter )
                    board = parent;
                  dst_contour->v_prev = (CvSeq*)parent;
                  cvSeqPush( root, &dst_contour );
              }
          }
      }
  }

  // finish contour retrieving 
  cvEndFindContours( &scanner );

  if (root->total < 1)
    return 0;
  // Get memory
  *out_ppQuads = (DQuadInfo *)cvAlloc(root->total * sizeof(DQuadInfo));
  *out_ppCorners = (DCornerInfo *)cvAlloc(root->total * 4 * sizeof(DCornerInfo));


  memset(*out_ppCorners, 0, root->total * 4 * sizeof(DCornerInfo));
  memset(*out_ppQuads, 0, root->total * sizeof(DQuadInfo));

  if (!*out_ppQuads || !*out_ppCorners)
    return -1;


  // Create array of quads structures 
  for( idx = 0; idx < root->total; idx++ )
  {
    src_contour = *(CvSeq**)cvGetSeqElem( root, idx );
    int i, total = src_contour->total;

    if( src_contour->v_prev != (CvSeq*)board )
        continue;

    cvStartReadSeq( src_contour, &reader, 0 );

    // Reset group ID
    (*out_ppQuads)[iQuadCount].m_iGroup = -1;

    assert( total == 4 );


#ifdef DEBUG_DRAW
    // DEBUG output     
    if (out_pDebugImage)
      cvDrawContours(out_pDebugImage, src_contour,
                     CV_RGB(255, 255, 255), CV_RGB(255, 255, 255),
                     0);
#endif //DEBUG_DRAW

    for( i = 0; i < total; i++ )
    {
      CvPoint pt;

      CV_READ_SEQ_ELEM( pt, reader );
    
      (*out_ppCorners)[iQuadCount * 4 + i].m_vPoint = cvPoint2D32f(pt.x, pt.y);
      (*out_ppQuads)[iQuadCount].m_pCorners[i] = &(*out_ppCorners)[iQuadCount * 4 + i];
    }

    iQuadCount++;
  }

  // release storages 
  cvReleaseMemStorage( &storage1 );

  return iQuadCount;
}

#ifdef DEBUG_DRAW

//=====================================================================================

void DrawQuad(CvMat *out_pImage, DQuadInfo *in_pQuad, unsigned char in_ucColor, int in_iThick)
{
  cvLine(out_pImage, 
           cvPointFrom32f(in_pQuad->m_pCorners[0]->m_vPoint),
           cvPointFrom32f(in_pQuad->m_pCorners[1]->m_vPoint), 
           in_ucColor, in_iThick);
  cvLine(out_pImage, 
           cvPointFrom32f(in_pQuad->m_pCorners[1]->m_vPoint),
           cvPointFrom32f(in_pQuad->m_pCorners[2]->m_vPoint), 
           in_ucColor, in_iThick);
  cvLine(out_pImage, 
           cvPointFrom32f(in_pQuad->m_pCorners[2]->m_vPoint),
           cvPointFrom32f(in_pQuad->m_pCorners[3]->m_vPoint), 
           in_ucColor, in_iThick);
  cvLine(out_pImage, 
           cvPointFrom32f(in_pQuad->m_pCorners[3]->m_vPoint),
           cvPointFrom32f(in_pQuad->m_pCorners[0]->m_vPoint), 
           in_ucColor, in_iThick);
}

//=====================================================================================

void ReleaseTmpImages()
{
  for (int i = 0; i < DEBUG_IMG_NUM; i++)
    cvReleaseImage( &g_ppDebugImages[i] );
}

#endif //DEBUG_DRAW
